from logging import Logger
from typing import Optional, Sequence, TYPE_CHECKING

import numpy as np
import pandas as pd

from truera.client.column_info import validate_column_input
from truera.client.util import workspace_validation_utils
# pylint: disable=no-name-in-module,no-member
from truera.protobuf.public import metadata_message_types_pb2 as md_pb

# pylint: enable=no-name-in-module
if TYPE_CHECKING:
    from truera.client.ingestion import ModelOutputContext
    from truera.client.ingestion import ColumnSpec

from truera.client.ingestion.constants import DEFAULT_MAX_EMBEDDINGS_LENGTH
from truera.client.ingestion.constants import DEFAULT_MAX_TOKEN_LENGTH
from truera.client.ingestion.constants import PROD_DATA_SPLIT_TYPE
from truera.client.ingestion.constants import SPECIAL_COLUMN_NAMES


def validate_column_spec_and_model_output_context(
    column_spec: 'ColumnSpec',
    model_output_context: 'ModelOutputContext',
    *,
    split_name: str,
    existing_models: Sequence[str],
    existing_splits: Sequence[str],
    project_influence_type: str,
    project_score_type: str,
    target_split_metadata=md_pb.DataSplitMetadata
):

    # Validate columns
    if not column_spec.id_col_name:
        raise ValueError("`id_col_name` is required in `column_spec`.")
    for column_type, column_names in column_spec.to_dict().items():
        if column_names:
            for column_name in validate_column_input(column_names):
                if column_name in SPECIAL_COLUMN_NAMES:
                    raise ValueError(
                        f"Column name '{column_name}' from `{column_type}` can not be used as it is reserved for use by TruEra services."
                    )

    # Validate target split
    is_appending = split_name in existing_splits and (
        column_spec.pre_data_col_names or column_spec.post_data_col_names
    )
    if is_appending:
        if target_split_metadata.split_type != PROD_DATA_SPLIT_TYPE:
            raise ValueError(
                f"Data split '{split_name}' already exists. Appending data is supported only for production monitoring data splits created by add_production_data_split."
            )
        if not column_spec.timestamp_col_name:
            raise ValueError(
                "Production monitoring data requires `timestamp_col_name` in `column_spec`."
            )

    # Validate model_output_context
    if column_spec.prediction_col_names or column_spec.feature_influence_col_names:
        if not model_output_context:
            raise ValueError(
                "`model_output_context` is required when ingesting predictions or feature influences."
            )

        # Validate model
        if not model_output_context.model_name:
            raise ValueError(
                "`model_name` is required in model output context when ingesting predictions or feature influences."
            )
        if model_output_context.model_name not in existing_models:
            raise ValueError(
                f"Model '{model_output_context.model_name}' does not exist among available models: {existing_models}"
            )

        # Validate score_type
        if not model_output_context.score_type:
            raise ValueError(
                f"`score_type` is required in model output context when ingesting predictions or feature influences."
            )
        # TODO (corey): need to sort out project score types and split score types in public API ingestion + quickstart nbs
        # elif model_output_context.score_type != project_score_type:
        #     raise ValueError(
        #         f"Got `score_type={model_output_context.score_type}` in model output context but project score type is {project_score_type}."
        #     )

        if column_spec.feature_influence_col_names:
            if not workspace_validation_utils.is_gradient_influence_type(
                model_output_context.influence_type
            ):
                # Validate background_split_name
                if not model_output_context.background_split_name:
                    raise ValueError(
                        "`background_split_name` is required in model output context when ingesting feature influences."
                    )
                if model_output_context.background_split_name not in existing_splits:
                    raise ValueError(
                        f"Data split '{model_output_context.background_split_name}' does not exist among available splits: {existing_splits}"
                    )

            # Validate influence_type
            if not model_output_context.influence_type:
                raise ValueError(
                    "`influence_type` is required in model output context when ingesting feature influences."
                )
            workspace_validation_utils.validate_influence_type_str_for_virtual_model_upload(
                model_output_context.influence_type, project_influence_type
            )


def validate_tokens(tokens: pd.Series, logger: Optional[Logger] = None):
    lengths = tokens.map(len)
    if logger is not None and not lengths.map(
        lambda x: x < DEFAULT_MAX_TOKEN_LENGTH
    ).all():
        logger.warning(
            f"Got tokens length of {lengths.max()} exceeds default limit of {DEFAULT_MAX_TOKEN_LENGTH}. This may result in performance issues."
        )
    if not tokens.map(lambda x: len(np.array(x).shape) == 1).all():
        raise ValueError(
            f"Tokens for each record must be a 1-dimensional array."
        )
    if not tokens.map(
        lambda x: np.issubdtype(np.array(x).dtype, "U") or np.
        issubdtype(np.array(x).dtype, "S")
    ).all():
        raise ValueError(f"Tokens for each record must be string types.")


def validate_embeddings(embeddings: pd.Series, logger: Optional[Logger] = None):
    lengths = embeddings.map(len)
    if len(lengths.unique()) > 1:
        raise ValueError(
            f"Embeddings must be of the same length, but got lengths: {lengths.unique().tolist()}."
        )
    if logger is not None and lengths.unique(
    )[0] > DEFAULT_MAX_EMBEDDINGS_LENGTH:
        logger.warning(
            f"Embeddings length of {lengths.unique()[0]} exceeds default limit of {DEFAULT_MAX_EMBEDDINGS_LENGTH}. This may result in performance issues."
        )
    if not embeddings.map(lambda x: len(np.array(x).shape) == 1).all():
        raise ValueError(
            f"Embeddings for each record must be a 1-dimensional array."
        )
    if not embeddings.map(
        lambda x: np.issubdtype(np.array(x).dtype, np.integer) or np.
        issubdtype(np.array(x).dtype, np.floating)
    ).all():
        raise ValueError(f"Embeddings for each record must be numeric.")


def validate_dataframe(
    data: pd.DataFrame,
    column_spec: 'ColumnSpec',
    input_type: str,
    logger: Optional[Logger] = None
):
    # Validate data against columns
    for column_type, column_names in column_spec.to_dict().items():
        if column_names:
            for column_name in validate_column_input(column_names):
                if column_name not in data:
                    raise ValueError(
                        f"Column name '{column_name}' was provided though `{column_type}`, but not found in data."
                    )
    if logger is not None:
        column_spec_columns = column_spec.get_all_columns()
        for column in data.columns:
            if column not in column_spec_columns:
                logger.warning(
                    f"Column '{column}' found in `data`, but will be ignored as it is not specified in `column_spec`."
                )

    # Validate id column is string
    try:
        data[column_spec.id_col_name].astype("string")
    except Exception as exc:
        raise ValueError(
            f"`id_col_name` '{column_spec.id_col_name}' column must be convertible into string type. Could not convert column to string; error: {exc}"
        )

    # Validate timestamp is datetime
    if column_spec.timestamp_col_name:
        try:
            pd.to_datetime(data[column_spec.timestamp_col_name])
        except Exception as exc:
            raise ValueError(
                f"`timestamp_col_name` column '{column_spec.timestamp_col_name}' must have valid datetime format. Could not convert column to datetime; "
                f"error: {exc}"
            )

    if column_spec.tokens_col_name:
        if input_type != "text":
            raise ValueError(
                f"`tokens_col_name` '{column_spec.tokens_col_name}' is only supported for text input."
            )
        validate_tokens(data[column_spec.tokens_col_name], logger=logger)
    if column_spec.embeddings_col_name:
        if input_type != "text":
            raise ValueError(
                f"`embeddings_col_name` '{column_spec.embeddings_col_name}' is only supported for text input."
            )
        validate_embeddings(
            data[column_spec.embeddings_col_name], logger=logger
        )

    # Validate feature influence columns for non-NLP ColumnSpec
    if column_spec.feature_influence_col_names:
        for col_name in column_spec.feature_influence_col_names:
            if input_type == "tabular":
                if not pd.api.types.is_float_dtype(data[col_name].dtype):
                    raise ValueError(
                        f"`feature_influence_data` column '{col_name}' must have float dtype, but was '{data[col_name].dtype}'"
                    )
            elif input_type == "text":
                if data[col_name].apply(
                    lambda x: not (
                        pd.api.types.is_array_like(x) or
                        isinstance(x, Sequence)
                    )
                ).any():
                    raise ValueError(
                        f"`feature_influence_data` column '{col_name}' must contain only array-like elements"
                    )
            else:
                raise ValueError(f"Invalid `input_type` specified {input_type}")
